import aiohttp
import pytz
from datetime import datetime
import os


# 10년치 데이터 종목당 180포인트 소모

cols1 = [
    'M111000',
    'M111100',
    'M111600',
    'M113000',
    'M113100',
    'M113800',
    'M113900',
    'M113500',
    'M115000',
    'M115020',
    'M115100',
    'M115500',
    'M115600',
    'M115850',
    'M121000',
    'M121100',
    'M121200',
    'M121300',
    'M121500',
    'M121540',
]

cols2 = [
    'M121550',
    'M121570',
    'M121580',
    'M122100',
    'M122500',
    'M122700',
    'M122710',
    'M122720',
    'M131000',
    'M132020',
    'M131025',
    'M132000',
    'M132010',
    'M133000',
    'M134000',
    'M134010',
    'M123500',
    'M123600',
    'M123900',
    'M123955',
]


async def get_performance(code, is_quarter: bool = True):
    """
    https://www.fnspace.com/Api/FinanceApi?key=sample&format=json&code=A005930&item='M111000',
    조회결과
    {
        key: string,
        format: string, (json / xml)
        success: string, (true / false),
        errcd: string, (성공일 경우""),
        errmsg: string,
        code: string,
        item: string,
        consolgb: string,
        annualgb: string,
        accdategb: string,
        fraccyear: string,
        toaccyear: string,
        dataset: object {
            CODE: string,
            NAME: string,
            DATA: object {
                DATE: string, # ex) 201509
                YYMM: string, # 결산 년월
                FS_YEAR: string, # 회계 년
                FS_MONTH: string, # 회계 월
                FS_QTR: string, # 회계 구분 ex) 3Q
                MAIN_TERM:string, # 주제무재표 구분 ex: 연결
                'M111000',
                'M111100',
                'M111600',
                ...
            }
        }
    }
    """
    print('fetch from fnguide', code)

    def validate(result):
        return (result is not None and
                'dataset' in result and
                len(result['dataset']) != 0 and
                'DATA' in result['dataset'][0])

    def merge(prev, row):
        data = prev['dataset'][0]['DATA']
        for prev_row in data:
            if (row['FS_YEAR'] == prev_row['FS_YEAR'] and
                    row['FS_MONTH'] == prev_row['FS_MONTH'] and
                    row['FS_QTR'] == prev_row['FS_QTR']):
                for key in row.keys():
                    if key.startswith('M') and key[1:].isnumeric():
                        prev_row[key] = row[key]

    url = 'https://www.fnspace.com/Api/FinanceApi'
    dt = datetime.now().astimezone(pytz.timezone("Asia/Seoul"))
    params = {
        'key': os.getenv('FN_GUIDE_KEY'),
        'code': code.upper(),
        'format': 'json',
        # 회계기준. 주재무제표(M)/연결(C)/별도(I) - 연결은 자본적 결합회사까지 합산
        'consolgb': 'M',
        # 연간(A)/분기(QQ)/분기누적(QY)
        'annualgb': 'QQ' if is_quarter else 'A',
        # 컨센서스 결산년월 선택 기준. Calendar(C)/Fiscal(F)
        'accdategb': 'C',
        'item': cols1,
        # 조회 시작 결산년월, toaccyear 에서 10년까지만 지원, 현재 년에서 -9
        'fraccyear': dt.year - 9,
        # 조회 종료 결산년월
        'toaccyear': dt.year
    }

    async with aiohttp.ClientSession() as session:
        d = None
        async with session.get(url, params=params) as response:
            # 존재하는 코드이나 값이 없으면, success: 'true' 이나 dataset: []
            if response.status == 200:
                d = await response.json()
        
        if validate(d):
            params['item'] = cols2
            async with session.get(url, params=params) as response:
                # 존재하는 코드이나 값이 없으면, success: 'true' 이나 dataset: []
                if response.status == 200:
                    result = await response.json()
                    if validate(result):
                        data = result['dataset'][0]['DATA']
                        for row in data:
                            merge(d, row)
        else:
            print(code, 'validate failed')
            return None
                            
        if d is not None:
            return d['dataset'][0]['DATA']
        return None



# -*- coding: utf-8 -*-
import re
import requests
from time import sleep
from fake_useragent import UserAgent
import datetime
import functools
import tempfile
from urllib.parse import unquote
import os


# Default Time-To-Live
CACHE_DEFAULT_TTL = datetime.timedelta(hours=1)


def cache(ttl=CACHE_DEFAULT_TTL):
    """ TTL Cache Decorator

    Parameters
    ----------
    ttl: datetime.timedelta
        Time-To-Live

    Returns
    -------
    function
        Wrapped Function
    """
    def wrap(func):
        cached = {}

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            now = datetime.datetime.now()
            # see lru_cache for fancier alternatives
            key = tuple(args), frozenset(kwargs.items())
            if key not in cached or now - cached[key][0] > ttl:
                value = func(*args, **kwargs)
                cached[key] = (now, value)
            return cached[key][1]
        return wrapped
    return wrap


@cache()
def get_user_agent():
    """ Return user-agent
    Returns
    -------
    str
        user-agent
    """
    ua = UserAgent()
    agent = ua.chrome
    return str(agent)


def query_to_regex(query):
    """ query to regular expression

    Parameters
    ----------
    query: str or list of str
        query

    Returns
    -------
    Pattern object
        regular expression
    """
    if isinstance(query, str):
        regex = re.compile(query, re.IGNORECASE)
    elif isinstance(query, list):
        pattern = '(' + '|'.join(query) + ')'
        regex = re.compile(pattern, re.IGNORECASE)
    else:
        raise TypeError('Invalid query type')
    return regex


class Request(object):
    """HTTP 요청을 보내는 클래스

    HTTP 요청을 위해 사용되는 클래스입니다.
    User-Agent 및 Cookies 관련 정보를 저장하고 있습니다.

    Attributes
    ---------
    s: Session
        Requests Session
    delay: float
        Delay for repeat delay, Default: 1s

    """
    def __init__(self):
        self.s = requests.Session()
        self.update_user_agent()
        # 분당 1000회 이상 자체적으로 24시간 IP차단
        # IP 차단 방지 위해 delay 0.1s -> 0.2s
        self.delay = 0.2

    def update_user_agent(self, force: bool = False):
        """ Update User-Agent

        Parameters
        ----------
        force: bool
            Force update
        """
        if force:
            ua = UserAgent()
            agent = ua.chrome
            user_agent = str(agent)
        else:
            user_agent = get_user_agent()
        self.s.headers.update({'user-agent': user_agent})

    def set_proxies(self, proxies: dict = None):
        """ Set proxies

        Parameters
        ----------
        proxies: dict
            proxies
        """
        if proxies is not None:
            import copy
            self.s.proxies = copy.deepcopy(proxies)

    def set_delay(self, second: float = None):
        """ Set delay

        Parameters
        ----------
        second: float
            delay for repeat
        """
        self.delay = second

    def request(self,
                url: str,
                method: str = 'GET',
                payload: dict = None,
                referer: str = None,
                stream: bool = False,
                timeout: int = 120):
        """ send http requests

        Parameters
        ----------
        url: str
            URL
        method: str, optional
            GET, OPTIONS, POST, PUT, PATCH or DELETE
        payload: dict, optional
            Request parameters
        referer: str, optional
            Temporary referer
        stream: bool, optional
            Stream optional, default False
        timeout: int, optional
            default 120s

        Returns
        -------
        requests.Response
            Response
        """
        headers = self.s.headers
        if referer is not None:
            headers['referer'] = referer

        """
        noqa: E501
        Session-level state such as cookies will not get 
        applied to your request.
        To get a PreparedRequest with that state applied,
        replace the call to Request.prepare() with
          a call to Session.prepare_request()
        """
        req = requests.Request(
            method,
            url=url,
            params=payload,
            headers=headers
        )
        prepped = self.s.prepare_request(req)
        resp = self.s.send(prepped, stream=stream, timeout=timeout)
        if self.delay is not None:
            sleep(self.delay)
        return resp

    def get(self, url: str,
            payload: dict = None,
            referer: str = None,
            stream: bool = False,
            timeout: int = 120):
        """
        request get method
        """
        return self.request(
            url=url,
            method='GET',
            payload=payload,
            referer=referer,
            stream=stream,
            timeout=timeout
        )

    def post(self, url: str,
             payload: dict = None,
             referer: str = None,
             stream: bool = False,
             timeout: int = 120):
        """ 
        request post method
        """
        return self.request(
            url=url,
            method='POST',
            payload=payload,
            referer=referer,
            stream=stream,
            timeout=timeout
        )

    def download(self,
                 url: str,
                 method: str = 'GET',
                 payload: dict = None,
                 referer: str = None,
                 timeout: int = 120) -> dict:
        """
        Download file from url and return file uri
        """
        r = self.request(
            url=url, method=method, payload=payload,
            referer=referer, stream=True, timeout=timeout)

        # Check validity
        headers = r.headers.get('Content-Disposition')
        if headers is None or not re.search('attachment', headers):
            raise FileNotFoundError('target does not exist')

        block_size = 8192

        # Extract filename
        extracted_filename = unquote(
            re.findall(r'filename="?([^"]*)"?', headers)[0]
        )
        temp_dir = tempfile.mkdtemp()
        file_path = os.path.join(temp_dir, extracted_filename)
        
        with open(file_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=block_size):
                if chunk is not None:
                    f.write(chunk)
        r.close()
        return file_path


# Request object
request = Request()



from enum import Enum
from typing import List


class MarketType(str, Enum):
    KOSPI = 'stockMkt'
    KOSDAQ = 'kosdaqMkt'
    KONEX = 'konexMkt'


class KrxCorp:
    def __init__(
        self,
        sectors: List[str] = [],
        products: List[str] = [],
        market_type: MarketType = MarketType.KOSPI,
        listed_date: str = '',
        name: str = '',
        stock_code: str = '',
        corp_code: str = '',
        update_date: str = ''
    ):
        self.sectors = sectors
        self.products = products
        self.market_type = market_type
        self.listed_date = listed_date
        self.name = name
        self.stock_code = stock_code
        self.halt_reason = ''
        self.is_halt = False
        self.corp_code = corp_code
        self.update_date = update_date

    @property
    def short_market_type(self):
        if self.market_type == MarketType.KOSPI:
            return 'Y'
        elif self.market_type == MarketType.KOSDAQ:
            return 'K'
        elif self.market_type == MarketType.KONEX:
            return 'N'
        return 'E'

    def __str__(self):
        return f'stock code({self.stock_code}) corp code({self.corp_code})' \
               f' name({self.name}) sectors({self.sectors})' \
               f' is_halt({self.is_halt}) reason({self.halt_reason})'



# -*- coding: utf-8 -*-
import re
from urllib.parse import urljoin

from .auth import get_api_key
from .status import check_status
from ..common.file import unzip_xbrl
from ..common.webrequest import request


# corp_code check regular expression
corp_code_checker = re.compile(r'^([0-9]{8})(,[0-9]{8})*$')
# bsns_year check regular expression
bsns_year_checker = re.compile(r'^[0-9]{4}$')
# reprt_code check regular expression
reptr_code_checker = re.compile(r'^1101[1-4]$')
# date check regular expression
date_checker = re.compile(r'^[0-9]{8}$')
# fs_div check regular expression
fs_div_checker = re.compile(r'CFS|OFS', re.IGNORECASE)
# sj_div check list
sj_div_checker = (
    'BS1', 'BS2', 'BS3', 'BS4',
    'IS1', 'IS2', 'IS3', 'IS4',
    'CIS1', 'CIS2', 'CIS3', 'CIS4',
    'DCIS1', 'DCIS2', 'DCIS3', 'DCIS4',
    'DCIS5', 'DCIS6', 'DCIS7', 'DCIS8',
    'CF1', 'CF2', 'CF3', 'CF4',
    'SCE1', 'SCE2'
)


def api_request(
        path: str,
        corp_code: str = None,
        bsns_year: str = None,
        reprt_code: str = None,
        bgn_de: str = None,
        end_de: str = None,
        fs_div: str = None,
        sj_div: str = None,
) -> dict:
    """ API Request Helper

    Parameters
    ----------
    path: str
        API Path
    corp_code: str
        Corporation Code (8 digits)
    bsns_year: str, optional
        Year (4 digits)
    reprt_code: str, optional
        Report code( Q1: 11013, half: 11012, Q3: 11014, Annual: 11011)
    bgn_de: str, optional
        Searching Start date (YYYYMMDD)
    end_de: str, optional
        Searching End date (YYYYMMDD)
    fs_div: str, optional
        CFS:Consolidated Financial Statements,
        OFS:Separate financial statements
    sj_div: str, optional
        Classification of financial statements
        (https://opendart.fss.or.kr/guide/detail.do?apiGrpCd=DS003&apiId=2020001)
    api_key: str, optional
        DART_API_KEY(If not set, use environment variable DART_API_KEY)
    Returns
    -------
    dict
        API Request Result
    """

    if corp_code and corp_code_checker.search(corp_code) is None:
        raise ValueError('corp_code must be 8 digits')

    if bsns_year and bsns_year_checker.search(bsns_year) is None:
        raise ValueError('bsns_year must be 4 digits')

    if reprt_code and reptr_code_checker.search(reprt_code) is None:
        raise ValueError('invalid reprt_code')

    if bgn_de and date_checker.search(bgn_de) is None:
        raise ValueError('invalid bgn_de')

    if end_de and date_checker.search(end_de) is None:
        raise ValueError('invalid end_de')

    if fs_div:
        if fs_div_checker.search(fs_div) is None:
            raise ValueError('fs_div must be CFS or OFS')
        fs_div = fs_div.upper()

    if sj_div:
        if sj_div not in sj_div_checker:
            raise ValueError('invalid sj_div')

    # Open DART Base URL
    base = 'https://opendart.fss.or.kr/'
    # Request URL
    url = urljoin(base, path)

    # Get DART_API_KEY
    api_key = get_api_key()

    # Set payload
    payload = {
        'crtfc_key': api_key,
        'corp_code': corp_code,
        'bsns_year': bsns_year,
        'reprt_code': reprt_code,
        'bgn_de': bgn_de,
        'end_de': end_de,
        'fs_div': fs_div,
        'sj_div': sj_div,
    }

    payload = {k: v for k, v in payload.items() if v is not None}

    # Request Data
    resp = request.get(url=url, payload=payload)

    # Convert Response to json
    dataset = resp.json()

    # Status Code Check
    check_status(**dataset)
    
    return dataset


def download_xbrl(path: str, rcept_no: str, reprt_code: str = None) -> str:
    """ XBRL 파일 다운로드

    Parameters
    ----------
    path: str
        Download Path
    rcept_no: str
        접수번호
    reprt_code: str, optinal
        1분기보고서 : 11013 반기보고서 : 11012 3분기보고서 : 11014 사업보고서 : 11011

    Returns
    -------
    str
        xbrl file path

    """
    url = 'https://opendart.fss.or.kr/api/fnlttXbrl.xml'

    payload = {
        'crtfc_key': get_api_key(),
        'rcept_no': rcept_no,
        'reprt_code': reprt_code
    }

    # Request Download
    file_uri = request.download(url=url, payload=payload)
    # Unzip File in User Cache Folder
    return unzip_xbrl(path, file_uri)



from arelle import ModelXbrl, XbrlConst
from dateutil.relativedelta import relativedelta
import re
import pandas as pd

from .util import str_to_regex

from .helper import (
    cls_datetime_check,
    cls_label_check,
    cls_merge_type,
    flatten,
    generate_df_columns,
    generate_df_rows,
    get_label_list,
    get_max_depth,
    get_title,
    get_value_from_dataset
)


class XbrlTable:
    """ XBRL Table

        XBRL 파일에서 추출된 데이터를 기반으로 재무제표에 관한 정보를 담고 있는 클래스

       Attributes
       ----------
       parent: str
           로드한 파일 이름
       xbrl: ModelXbrl
           arelle Xbrl 클래스

       """
    def __init__(self, parent, xbrl: ModelXbrl, code, definition, uri):
        self.parent = parent
        self.code = code
        self.definition = definition
        self.uri = uri
        self._xbrl = xbrl
        self._facts = None
        self._dataset = None
        self._cls = None
        self._labels = None

    @property
    def facts(self):
        """list of modelFact: """
        if self._facts is None:
            arcrole = XbrlConst.parentChild
            relation = self._xbrl.relationshipSet(arcrole, self.uri)
            facts = []
            for fact in self._xbrl.facts:
                if relation.fromModelObject(fact.concept) \
                        or relation.toModelObject(fact.concept):
                    facts.append(fact)
            self._facts = facts
        return self._facts

    @property
    def dataset(self):
        """dict of modelFact: """
        if self._dataset is None:
            dataset = dict()
            for fact in self.facts:
                object_id = fact.context.objectId()
                if dataset.get(object_id) is None:
                    dataset[object_id] = []
                dataset[object_id].append(fact)

            self._dataset = dataset
        return self._dataset

    @property
    def cls(self):
        """classification 반환"""
        if self._cls is None:
            self._get_cls()
        return self._cls

    def cls_filter(self, start_dt=None, end_dt=None, label=None):
        """ classification 필터링 함수

        Parameters
        ----------
        start_dt: str
            검색 시작 일자
        end_dt: str
            검색 종료 일자
        label: str
            포함할 label 명

        Returns
        -------
        list of cls
            필터된 classification
        """
        return [item for item in self.cls
                if cls_datetime_check(item, start_dt, end_dt) and cls_label_check(item, label)]

    def _get_cls(self):
        """ classification 정보 추출 함수"""
        contexts = set()
        for data in self.facts:
            context = data.context
            contexts.add(context)

        cls = list()
        for context in contexts:
            object_id = context.objectId()

            # data가 없을때 무시
            if len(self.dataset[object_id]) < 1:
                continue

            instant_datetime = None
            start_datetime = None
            end_datetime = None
            if context.isInstantPeriod is True:
                instant_datetime = context.instantDatetime - relativedelta(days=1)

            else:
                start_datetime = context.startDatetime
                end_datetime = context.endDatetime - relativedelta(days=1)

            label = dict()
            dims = context.qnameDims
            if len(dims) > 0:
                for dimQname in sorted(dims.keys(), key=lambda d: str(d), reverse=True):
                    dim_value = dims[dimQname]
                    ko = dim_value.member.label(lang='ko')
                    ko = re.sub(r'\[.*?\]', '', ko)
                    en = dim_value.member.label(lang='en')
                    en = re.sub(r'\[.*?\]', '', en)
                    label[dimQname] = {
                        'ko': ko,
                        'en': en
                    }
            _cls = {
                'cls_id': object_id,
                'instant_datetime': instant_datetime,
                'start_datetime': start_datetime,
                'end_datetime': end_datetime,
                'label': label
            }
            cls.append(_cls)
        cls.sort(key=lambda x: x.get('instant_datetime') or x.get('start_datetime'), reverse=True)
        self._cls = cls
        return self._cls

    @property
    def labels(self):
        """labels 반환"""
        if self._labels is None:
            self._labels = {}
            arcrole = XbrlConst.parentChild
            relationship_set = self._xbrl.relationshipSet(arcrole, self.uri)
            for root_concept in relationship_set.rootConcepts:
                labels = get_label_list(relationship_set, root_concept)
                self._labels = {**self._labels, **labels}
        return self._labels

    def to_DataFrame(self, cls=None, lang='ko', start_dt=None, end_dt=None,
                     label=None, show_abstract=False, show_class=True, show_depth=10,
                     show_concept=True, separator=True):
        """ Pandas DataFrame으로 변환하는 함수

        Parameters
        ----------
        cls: dict, optional
            classification
        lang: str, optional
            'ko' 한글 or 'en' 영문
        start_dt: str, optional
            검색 시작 일자
        end_dt: str, optional
            검색 종료 일자
        label: str, optional
            Column Label에 포함될 단어
        show_abstract: bool, optional
            abtract 표시 여부
        show_class: bool, optional
            class 표시여부
        show_depth: int, optional
            class 표시 깊이
        show_concept: bool, optional
            concept_id 표시 여부
        separator: bool, optional
            숫자 첫단위 표시 여부

        Returns
        -------
        DataFrame
            재무제표 DataFrame
        """
        if cls is None:
            cls = self.cls_filter(start_dt, end_dt, label)
        cls = cls_merge_type(cls)
        depth = get_max_depth(self.labels, show_abstract=show_abstract)
        depth = depth if depth < show_depth else show_depth

        table = self.parent.get_table_by_code('d999004')
        unit = get_value_from_dataset(table.cls, table.dataset, 'dart-gcd_EntityReportingCurrencyISOCode')

        definition = self.definition + ' (Unit: {})'.format(unit[0])
        columns = generate_df_columns(definition, cls, depth, lang,
                                      show_concept=show_concept, show_class=show_class)

        if separator:
            pd.options.display.float_format = '{:,}'.format
        else:
            pd.options.display.float_format = '{:}'.format
        df = pd.DataFrame(columns=columns)

        rows = generate_df_rows(self.labels, cls, self.dataset, depth, lang=lang,
                                show_abstract=show_abstract, show_concept=show_concept, show_class=show_class)
        data = flatten(rows)
        for idx, r in enumerate(data):
            df.loc[idx] = r

        regex_pass = str_to_regex('concept_id OR label_ko OR label_en OR class')
        df_count = df.count()
        drop_columns = []
        for key, count in df_count.items():
            if regex_pass.search(' '.join(key[1])):
                pass
            elif count <= 1:
                drop_columns.append(key)
        df = df.drop(drop_columns, axis=1)
        return df

    def get_value_by_concept_id(self, concept_id, start_dt=None, end_dt=None, label=None, lang='en'):
        """ concept_id을 이용하여 값을 찾아 주는 함수

        Parameters
        ----------
        concept_id: str
            재무제표 계정의 concept_id
        start_dt: str
            검색 시작 일자
        end_dt: str
            검색 종료 일자
        label: str
            검색 포함 label
        lang: str
            'ko' 한글 / 'en' 영문

        Returns
        -------
        dict of (str or float)
            { column 이름 : 값 }
        """
        cls = self.cls_filter(start_dt, end_dt, label)
        data = get_value_from_dataset(classification=cls, dataset=self.dataset, concept_id=concept_id)
        results = dict()
        for c, d in zip(cls, data):
            title = get_title(c, lang=lang)
            results[title] = d
        return results

    def __repr__(self):
        info = {
            'code': self.code,
            'definition': self.definition
        }
        return str(info)



from datetime import datetime
from typing import Union
import re


CURRENCY = {
    '원': 'KWR',
    '달러': 'USD',
    '엔': 'JPY',
}

str_or_datetime = Union[str, datetime]


def str_compare(str1: str, str2: str) -> bool:
    """문자열 비교

    Parameters
    ----------
    str1: str
        string
    str2: str
        string

    Returns
    -------
    bool
        문자열이 동일하다면 True
    """
    if not isinstance(str1, str) or not isinstance(str2, str):
        return False
    str1 = str1.strip().lower()
    str2 = str2.strip().lower()
    return str1 == str2


def get_datetime(date: str_or_datetime) -> datetime:
    """ 문자열을 datetime올 변환

    Parameters
    ----------
    date: str or datetime
        datetime 문자열

    Returns
    -------
    datetime
        변환된 datetime

    """
    if isinstance(date, str):
        return datetime.strptime(date, '%Y%m%d')
    elif isinstance(date, datetime):
        return date
    else:
        raise ValueError('Invalid datetime format')


def check_datetime(date: str_or_datetime,
                   start_date: str_or_datetime = None,
                   end_date: str_or_datetime = None) -> bool:
    """ Date가 올바른지 체크하는 함수

    Parameters
    ----------
    date: str or datetime
        체크할 값
    start_date: str or datetime
        시작 일자
    end_date: str or datetime
        종료 일자

    Returns
    -------
    bool
        Date가 start_date와 end_date 사이에 있는지 여부
    """
    date = get_datetime(date)
    if start_date is not None:
        start_date = get_datetime(start_date)
        if date < start_date:
            return False
    if end_date is not None:
        end_date = get_datetime(end_date)
        if date > end_date:
            return False
    return True


def str_unit_to_number_unit(str_unit: str) -> int:
    """ 통화 단위를 숫자로 변화

    Parameters
    ----------
    str_unit: str
        통화 단위

    Returns
    -------
    int
        환산값
    """
    str_unit = re.sub(r'\s+', '', str_unit)

    str_unit_to_unit = {
        '억원': 100000000,
        '천만원': 10000000,
        '백만원': 1000000,
        '십만원': 100000,
        '만원': 10000,
        '천원': 1000,
        '백원': 100,
        '십원': 10,
    }

    for k, v in CURRENCY.items():
        str_unit_to_unit[k] = 1
        str_unit_to_unit[v] = 1

    return str_unit_to_unit[str_unit]


def get_currency_str(unit: str) -> Union[str, None]:
    regex_str = ' OR '.join(CURRENCY.keys())
    str_unit = str_to_regex(regex_str).search(unit)
    if str_unit:
        str_unit = str_unit.group(0)
        return CURRENCY[str_unit]

    regex_str = ' OR '.join([v for _, v in CURRENCY.items()])
    str_unit = str_to_regex(regex_str).search(unit)
    if str_unit:
        return str_unit.group(0)

    return None


def str_to_regex(query):
    """ regular expression

    Parameters
    ----------
    query: str
        검색 문구

    Returns
    -------
    Pattern
        regular expression pattern
    """
    return re.compile(str_to_pattern(query))


def str_to_pattern(query):
    """ AND OR 등 연산자를 regular expression 표현으로 변경

    Parameters
    ----------
    query: str
        검색 문구

    Returns
    -------
    str
        검색 pattern
    """
    query = query.replace('(', ' ( ').replace(')', ' ) ').split()
    postfix = infix_to_postfix(query)

    stack = []
    for item in postfix:
        if is_operator(item):
            if item == 'AND':
                operand1 = stack.pop()
                operand2 = stack.pop()
                pattern = r'(?=.*{0})(?=.*{1})'.format(operand2, operand1)
            elif item == 'OR':
                operand1 = stack.pop()
                operand2 = stack.pop()
                pattern = r'({0}|{1})'.format(operand2, operand1)
            stack.append(pattern)
        else:
            stack.append(item)
    return stack.pop()


def is_operator(item):
    """  연산자 여부 검색

    Parameters
    ----------
    item : str
        단어

    Returns
    -------
    bool
        True: 연산자 / False: 일단단어

    """
    if item in ['AND', 'OR']:
        return True
    return False


def precedence(symbol):
    """ 연산자 우선수위

    Parameters
    ----------
    symbol: str
        연산자

    Returns
    -------
    int
        연산자 순위

    """
    if symbol in ['AND', 'OR']:
        return 1
    return 0


def infix_to_postfix(infix):
    """ infix를 postfix로 변환하는 함수

    Parameters
    ----------
    infix: str
        infix 형태로 표현된 문장

    Returns
    -------
    str
        postfix로 변횐된 문장
    """
    stack = []
    results = []
    parenthesis = 0
    for item in infix:
        if item == '(':
            parenthesis = parenthesis + 1
            stack.append(item)
        elif item == ')':
            parenthesis = parenthesis - 1
            if parenthesis < 0:
                raise SyntaxError('Missing left parentheses.')
            while stack[-1] != '(':
                v = stack.pop()
                results.append(v)
            stack.pop()
        elif is_operator(item):
            if len(stack) > 0:
                if precedence(stack[-1]) < precedence(item):
                    stack.append(item)
                else:
                    v = stack.pop()
                    stack.append(item)
                    results.append(v)
            else:
                stack.append(item)
        else:
            results.append(item)
    while len(stack) > 0:
        results.append(stack.pop())

    if parenthesis > 0:
        raise SyntaxError('Missing right parentheses.')

    return results



import logging
from datetime import datetime
from krxmarket import fnltt_singl_acnt_all


LOGGER = logging.getLogger(__name__)


def _fv(item: dict, key: str):
    value = item.get(key)
    return value


def get_performance(corp_code: str, start_year: int) -> list:
    """
    json structure
    [
        {
            'year': '2016',
            'quarter': 1,
            'apx_ms': xxxxxxxx
            data: {
                CF: [
                    {},
                    {}
                ],
                IS: {
                
                }
            }
        }
    ]
    """
    year = start_year
    current_year = datetime.now().year
    # 1분기보고서:11013, 반기보고서:11012, 3분기보고서:11014, 사업보고서 : 11011
    report_seq = ['11013', '11012', '11014', '11011']
    report_all = []

    while year <= current_year:
        for i, seq in enumerate(report_seq):
            result = {
                'year': str(year),
                'quarter': str(i + 1),
                'data': {}
            }
            try:
                report = fnltt_singl_acnt_all(corp_code, str(year), seq, 'CFS')
                if report['status'] == '000' and 'list' in report:
                    items = _fv(report, 'list')
                    for item in items:
                        sj_div = _fv(item, 'sj_div')
                        if sj_div not in result['data']:
                            result['data'][sj_div] = []
                        
                        result['data'][sj_div].append({
                            'account_nm': _fv(item, 'account_nm'),
                            'frmtrm_amount': _fv(item, 'frmtrm_amount'),
                            'thstrm_amount': _fv(item, 'thstrm_amount'),
                            'account_id': _fv(item, 'account_id'),
                            'rcept_no': _fv(item, 'rcept_no')
                        })
            except Exception as ex:
                LOGGER.warning('skip %d, quarter: %d, %s',
                               year, i + 1, str(ex))
                continue
                
            report_all.append(result)
        year += 1

    return report_all


if __name__ == '__main__':
    json_content = get_performance('00126380', 2022)
    import json
    with open('data.json', 'w') as f:
        json.dump(json_content, f)




import sys
import os
from typing import List, Union
from arelle import ModelXbrl, XbrlConst
import re
from pandas import DataFrame
from arelle import Cntlr
import pandas as pd

from .helper import consolidated_code_to_role_number, get_title
from .util import str_compare
from .xbrl_table import XbrlTable


def arelle_setup():
    if sys.platform == 'win32':
        pass
    elif sys.platform == 'darwin':
        arelle_app_dir = os.path.join(os.path.expanduser('~/Library/Application Support'), 'Arelle')
        if not os.path.exists(arelle_app_dir):
            os.makedirs(arelle_app_dir)
    else:
        arelle_app_dir = os.path.join(os.path.expanduser("~/.config"), "arelle")
        if not os.path.exists(arelle_app_dir):
            os.makedirs(arelle_app_dir)


class Xbrl:
    def __init__(self, file_uri: str):
        arelle_setup()
        self.xbrl: ModelXbrl = Cntlr.Cntlr(
            logFileName='temp.log').modelManager.load(file_uri)
        self._tables = None

    def get_fs(self):
        data = self.get_financial_statement()
        return data[0] if data else None

    def get_is(self):
        data = self.get_income_statement()
        if data:
            data = data[0] if len(data) > 1 else None
        return data

    def get_ci(self):
        data = self.get_income_statement()
        if data:
            data = data[1] if len(data) > 1 else data[0]
        return data

    def get_cf(self):
        data = self.get_cash_flows()
        return data[0] if data else None

    @property
    def tables(self):
        """list of Table: Table 리스트"""
        if self._tables is not None:
            return self._tables

        arcrole = XbrlConst.parentChild
        relationship = self.xbrl.relationshipSet(arcrole)
        tables = None
        if relationship is not None:
            tables = []
            for uri in relationship.linkRoleUris:
                role_types = self.xbrl.roleTypes.get(uri)

                if role_types is not None:
                    definition = (role_types[0].genLabel(lang='ko', strip=True)
                                  or role_types[0].definition or uri)
                else:
                    definition = uri

                role_code = re.search(r"\[(.*?)\]", definition)
                role_code = role_code.group(1) if role_code else None
                tables.append(XbrlTable(self, self.xbrl, role_code, definition, uri))
                
        self._tables = tables
        return tables

    def get_table_by_code(self, code: str) -> Union[XbrlTable, None]:
        for table in self.tables:
            if str_compare(table.code, code):
                return table
        return None

    def _to_info_DataFrame(self, code: str, lang: str = 'ko') -> DataFrame:
        table = self.get_table_by_code(code)
        return table.to_DataFrame(lang=lang, show_class=False, show_concept=False, separator=False)
    
    def get_document_information(self, lang: str = 'ko') -> DataFrame:
        return self._to_info_DataFrame('d999001', lang=lang)

    def get_entity_information(self, lang: str = 'ko') -> DataFrame:
        """
        공시 대상 정보
        """
        return self._to_info_DataFrame('d999004', lang=lang)

    def get_audit_information(self, lang: str = 'ko') -> DataFrame:
        """
        감사 정보
        """
        return self._to_info_DataFrame('d999003', lang=lang)

    def get_entity_address_information(self, lang: str = 'ko') -> DataFrame:
        return self._to_info_DataFrame('d999005', lang=lang)

    def get_author_information(self, lang: str = 'ko') -> DataFrame:
        return self._to_info_DataFrame('d999006', lang=lang)

    def get_financial_statement_information(self, lang: str = 'ko') -> DataFrame:
        """
        재무제표
        """
        return self._to_info_DataFrame('d999007', lang=lang)

    def get_financial_statement(self, separate: bool = False) -> Union[List[XbrlTable], None]:
        return self._get_statement('dart-gcd_StatementOfFinancialPosition', separate=separate)

    def get_income_statement(self, separate: bool = False) -> Union[List[XbrlTable], None]:
        return self._get_statement('dart-gcd_StatementOfComprehensiveIncome', separate=separate)

    def get_changes_in_equity(self, separate: bool = False) -> Union[List[XbrlTable], None]:
        return self._get_statement('dart-gcd_StatementOfChangesInEquity', separate=separate)

    def get_cash_flows(self, separate: bool = False) -> Union[List[XbrlTable], None]:
        return self._get_statement('dart-gcd_StatementOfCashFlows', separate=separate)

    def _get_statement(
        self,
        concept_id: str,
        separate: bool = False
    ) -> Union[List[XbrlTable], None]:
        """ Financial statement information 을 이용하여 제공되는 재무제표를 추출하는 함수

        Parameters
        ----------
        concept_id: str
            dart-gcd_StatementOfFinancialPosition: 재무상태표
            dart-gcd_StatementOfComprehensiveIncome: 포괄손익계산서
            dart-gcd_StatementOfChangesInEquity: 자본변동표
            dart-gcd_StatementOfCashFlows: 현금프름표
        separate: bool, optional
            True: 개별재무제표
            False: 연결재무제표

        Returns
        -------
        Table or None


        """
        table = self.get_table_by_code('d999007')
        if table is None:
            return None
        table_dict = table.get_value_by_concept_id(concept_id)
        compare_name = 'Separate' if separate else 'Consolidated'
        for keys, value in table_dict.items():
            for key in keys:
                title = ''.join(key)
                if re.search(compare_name, title, re.IGNORECASE):
                    code_list = consolidated_code_to_role_number(value, separate=separate)
                    tables = [self.get_table_by_code(code) for code in code_list]
                    return tables
        return None

    def exist_consolidated(self):
        """ 연결 재무제표 존재 여부를 확인하기 위한 함수

        Returns
        -------
        bool
            연결재무제표 존재시 True / 개별재무제표만 존재시 False
        """
        regex = re.compile(r'Consolidated', re.IGNORECASE)
        info_table = self.get_table_by_code('d999007')
        cls_list = info_table.cls
        for cls in cls_list:
            titles = get_title(cls, 'en')
            for title in titles:
                if isinstance(title, str):
                    if regex.search(title):
                        return True
                else:
                    if regex.search(' '.join(title)):
                        return True
        return False
    
    def get_period_information(self, lang: str = 'ko') -> DataFrame:
        """ 공시 문서 기간 정보

        Parameters
        ----------
        lang: str
            표시 언어 설정('ko': 한글, 'en': 영어)

        Returns
        -------
        DataFrame
            공시 문서 기간 정보
        """
        df = self._to_info_DataFrame('d999002', lang=lang)
        data = df[df.columns[2:]].iloc[3]
        data_set = [(key, data[key]) for key in data.keys()]
        new_columns = list(df.columns[:2]) + \
            [data[0] for data in sorted(data_set, key=lambda x: x[1], reverse=True)]
        new_columns = pd.MultiIndex.from_tuples(new_columns)
        return df[new_columns]


from datetime import datetime
import math
import re
import warnings
import pandas as pd

from .util import (
    check_datetime,
    get_currency_str,
    get_datetime,
    str_compare,
    str_to_regex,
    str_unit_to_number_unit
)


def get_label_list(relationship_set, concepts, relationship=None):
    """ XBRL의 label list를 변환하는 함수 """
    if relationship is None:
        if len(relationship_set.modelRelationships) > 0:
            preferred = relationship_set.modelRelationships[0].preferredLabel
        else:
            preferred = None
    else:
        preferred = relationship.preferredLabel

    label_ko = concepts.label(lang='ko') if preferred is None \
        else concepts.label(preferredLabel=preferred, lang='ko')
    label_en = concepts.label(lang='en') if preferred is None \
        else concepts.label(preferredLabel=preferred, lang='en')

    res = {
        'concept_id': concepts.id,
        'order': 1.0 if relationship is None else relationship.order,
        'label_ko': label_ko,
        'label_en': label_en,
        'isAbstract': concepts.isAbstract,
        'children': []
    }

    new_relationship_set = relationship_set.fromModelObject(concepts)
    if len(new_relationship_set) > 0:
        new_relationship_set.sort(key=lambda x: x.order)
        for rel in new_relationship_set:
            new_concepts = rel.viewConcept
            res['children'].append(get_label_list(relationship_set, new_concepts, relationship=rel))
    return res


def flatten(tuple_list, res=None):
    """ 2차원 list를 1차원 리스트로 변환하는 함수 """
    results = list() if res is None else res
    for data in tuple_list:
        if isinstance(data, list):
            flatten(data, results)
        else:
            results.append(data)
    return results


def get_max_depth(labels, show_abstract=False):
    """ class의 최대 깊이를 확인하는 함수"""
    max_depth = 0

    if len(labels['children']) == 0 and labels['isAbstract'] is True:
        if show_abstract is True:
            return 1
        else:
            return 0

    for label in labels['children']:
        depth = get_max_depth(label, show_abstract=show_abstract)
        if max_depth < depth:
            max_depth = depth

    return 1 + max_depth


def get_title(cls, lang='ko'):
    """ cls에서 column의 title을 생성하는 함수 """
    title = []

    if cls['instant_datetime'] is None:
        start_date = cls['start_datetime'].strftime('%Y%m%d')
        end_date = cls['end_datetime'].strftime('%Y%m%d')
        title.append('{}-{}'.format(start_date, end_date))
    else:
        instant_date = cls['instant_datetime'].strftime('%Y%m%d')
        title.append('{}'.format(instant_date))
    qlist = []
    for qname in cls['label']:
        qlist.append(cls['label'][qname][lang].strip())
    title.append(tuple(qlist))
    return tuple(title)


def get_datetime_and_name(title):
    """ Column Title에서 날짜와 Tilte 명을 추출하는 함수"""
    regx = re.search(r'\[(.*?)\]', title, re.IGNORECASE)
    result = None
    if regx is not None:
        result = dict()
        instant_datetime = None
        start_datetime = None
        end_datetime = None

        datetime = regx.group(1).split(',')
        if len(datetime) == 1:
            instant_datetime = get_datetime(datetime[0])
        elif len(datetime) == 2:
            start_datetime = get_datetime(datetime[0])
            end_datetime = get_datetime(datetime[1])
        else:
            raise ValueError('Invalid date')

        name = re.sub(r'\[{}\]'.format(regx.group(1)), '', title)
        name = name.strip()

        result['instant_datetime'] = instant_datetime
        result['instant_date'] = start_datetime
        result['start_date'] = end_datetime
        result['name'] = name
    return result


def get_value_from_dataset(classification, dataset, concept_id, label_ko=None):
    """ dataset에서 값을 추출하는 함수 """
    def str_to_float(val):
        try:
            return float(val)
        except ValueError:
            return val

    if isinstance(classification, dict):
        classification = [classification]

    # XBRL 내부 주당이익에서 발생하는 오류 수정을 위한 코드
    currency_unit = None
    if label_ko is not None:
        regex = re.compile(r'\(단위:(.*)\)')
        unit = regex.search(label_ko)
        if unit is not None:
            unit = unit.group(0)
            currency = get_currency_str(unit)
            if currency is not None:
                currency_unit = str_unit_to_number_unit(currency)

    results = list()
    added_title = list()
    for cls in classification:
        value = float('nan')
        for data in dataset[cls['cls_id']]:
            if str_compare(data.concept.id, concept_id):
                value = str_to_float(data.value)
                # XBRL 내부 주당이익에서 발생하는 오류 수정을 위한 코드
                if currency_unit is not None:
                    decimals = str_to_float(data.decimals)
                    # decimals이 없을 경우 0으로 처리
                    if math.isinf(decimals) or math.isnan(decimals):
                        decimals = 0
                    value = value * pow(10, decimals)
                    value = value * currency_unit
                break

        title = get_title(cls, 'en')
        if title in added_title:
            index = added_title.index(title)
            if not math.isnan(value):
                results[index] = value
        else:
            results.append(value)
            added_title.append(title)
    return results


def generate_df_columns(definition, classification, max_depth, lang='ko', show_concept=True, show_class=True):
    """ Table의 DataFrame 변환시 Column Title 생성을 위한 함수"""
    columns = [(definition, 'concept_id')] if show_concept else []
    columns += [(definition, 'label_ko'), (definition, 'label_en')]

    if show_class:
        for idx in range(max_depth):
            columns.append((definition, 'class{}'.format(idx)))

    if isinstance(classification, dict):
        classification = [classification]

    added_title = list()
    for cls in classification:
        title = get_title(cls, lang)
        key = get_title(cls, 'en')
        if key not in added_title:
            columns.append(title)
            added_title.append(key)
    return pd.MultiIndex.from_tuples(columns)


def generate_df_rows(labels, classification, dataset, max_depth,
                     lang="ko", parent=(), show_abstract=False,
                     show_concept=True, show_class=True):
    """ Table의 DataFrame으로 변환시 DataFrame의 Row 생성을 위한 함수"""
    lang_type = {
        'ko': 'label_ko',
        'en': 'label_en',
        'concepts': 'concept_id'
    }
    results = []
    new_parent = list(parent)
    new_parent.append(labels[lang_type[lang]])

    if show_class:
        if len(new_parent) > max_depth:
            return None

    row = None
    if not labels['isAbstract'] or show_abstract:
        row = [labels['concept_id']] if show_concept else []
        row += [labels['label_ko'], labels['label_en']]
        if show_class:
            for idx in range(max_depth):
                if idx < len(new_parent):
                    row.append(new_parent[idx])
                else:
                    row.append(None)
        row.extend(get_value_from_dataset(
            classification,
            dataset,
            labels['concept_id'],
            labels['label_ko']
        ))
        results.append(tuple(row))

    if len(labels['children']) > 0:
        for child in labels['children']:
            generated_row = generate_df_rows(
                child, classification, dataset, max_depth,
                lang=lang, parent=tuple(new_parent), show_abstract=show_abstract,
                show_concept=show_concept, show_class=show_class
            )
            if generated_row is not None:
                results.append(generated_row)
    else:
        return tuple(row) if row is not None else None

    return results


def consolidated_code_to_role_number(code, separate=False):
    """ 코드번호를 Role 번호로 변환하는 함수

    Parameters
    ----------
    code: str
        코드번호
    separate: bool, optional
        개별재무제표 여부

    Returns
    -------
    list of str
        Role 번호 리스트
    """
    consolidated_code = {
        'D1001': ['D210000'],
        'D1002': ['D220000'],
        'D2001': ['D431410'],
        'D2002': ['D431420'],
        'D2003': ['D432410'],
        'D2004': ['D432420'],
        'D2005': ['D310000', 'D410000'],
        'D2006': ['D310000', 'D420000'],
        'D2007': ['D320000', 'D410000'],
        'D2008': ['D320000', 'D420000'],
        'D2009': ['D310000'],
        'D2010': ['D320000'],
        'D3001': ['D610000'],
        'D4001': ['D510000'],
        'D4002': ['D520000'],
    }
    separated_code = {
        'D1001': ['D210005'],
        'D1002': ['D220005'],
        'D2001': ['D431415'],
        'D2002': ['D431425'],
        'D2003': ['D432415'],
        'D2004': ['D432425'],
        'D2005': ['D310005', 'D410005'],
        'D2006': ['D310005', 'D420005'],
        'D2007': ['D320005', 'D410005'],
        'D2008': ['D320005', 'D420005'],
        'D2009': ['D310005'],
        'D2010': ['D320005'],
        'D3001': ['D610005'],
        'D4001': ['D510005'],
        'D4002': ['D520005'],
    }
    return separated_code[code] if separate else consolidated_code[code]


def cls_datetime_check(cls, start_dt, end_dt):
    """ classification 의 시간이 시작일자와 종료일자 사이인지 체크하는 함수

    Parameters
    ----------
    cls: cls
        classification
    start_dt: str
        검색 시작 일자
    end_dt: str
        검색 종료 일자

    Returns
    -------
    bool
        시작일자와 종료일자 사이에 있을시 True / 아닐시 False

    """
    res = True
    instant_datetime = cls.get('instant_datetime')
    if instant_datetime:
        res = res and check_datetime(instant_datetime, start_dt, end_dt)
    start_datetime = cls.get('start_datetime')
    if start_datetime:
        res = res and check_datetime(start_datetime, start_dt, end_dt)
    end_datetime = cls.get('end_datetime')
    if end_datetime:
        res = res and check_datetime(end_datetime, start_dt, end_dt)
    return res


def cls_label_check(cls, query):
    """ classification label에 특정 단어가 포함된지 검색하는 함수

    Parameters
    ----------
    cls: cls
        classification
    query: str
        검색어

    Returns
    -------
    bool
        질의내용 포함시 True / 미포함시 False
    """
    if query is None:
        return True
    regex = str_to_regex(query)
    label = ''
    for qname in cls['label']:
        label = label + cls['label'][qname]['ko'] + cls['label'][qname]['en']
    if regex.search(label):
        return True
    return False


def cls_merge_type(classification):
    """ classification type이 2가지일 때 합쳐주는 함수

    Parameters
    ----------
    classification: cls
        classification 리스트

    Returns
    -------
    list of cls
        변환된 classification 리스트
    """

    cls_datetime = {}

    for cls in classification:
        end_datetime = cls.get('end_datetime')
        start_datetime = cls.get('start_datetime')
        if end_datetime is not None:
            cls_datetime[end_datetime] = start_datetime

    if len(cls_datetime) > 0:
        for cls in classification:
            instant_datetime = cls.get('instant_datetime')
            if instant_datetime:
                start_datetime = cls_datetime.get(instant_datetime)
                if start_datetime is None:
                    warnings_text = "start_datetime could not be found. " \
                                    "Therefore, start_datetime is set to January 01, 1900"
                    warnings.warn(warnings_text, RuntimeWarning)
                    start_datetime = datetime(year=1900, month=1, day=1)
                cls['instant_datetime'] = None
                cls['start_datetime'] = start_datetime

                cls['end_datetime'] = instant_datetime

    return classification



from ..common.exception import (
    APIKeyError, TemporaryLocked, NoDataReceived,
    OverQueryLimit, InvalidField, ServiceClose, UnknownError
)


def check_error(status):
    errors = {
        '000': None,
        '010': APIKeyError,
        '011': TemporaryLocked,
        '013': NoDataReceived,
        '020': OverQueryLimit,
        '100': InvalidField,
        '800': ServiceClose,
        '900': UnknownError,
    }
    return errors.get(status, UnknownError)


def check_status(**kwargs):
    """
    Open DART의 응답 결과중 status 값을 체크하는 함수
    Parameters
    ----------
    kwargs: dict
        DART 서버에서 전달받은 값

    Raises
    ------
    APIKeyError
        등록되지 않은 API 키에 의해 발생하는 오류
    TemporaryLocked
        오픈 API에 등록 되었으나, 일시적으로 사용 중지된 키에 의해 발생하는 오류
    OverQueryLimit
        요청 제한을 초과하였을때 발생하는 오류
    InvalidField
        요청한 필드에 부적절한 값이 있는 경우 발생하는 오류
    ServiceClose
        원할한 공시서비스를 위해 오픈 API 서비스가 중지 되었을때 발생하는 오류
    UnknownError
        정의되지 않은 오류
    """
    status = kwargs.get('status')
    err = check_error(status)
    if err is not None:
        msg = kwargs.get('message')
        raise err(msg)



from typing import Dict
import logging

from .common.types import KrxCorp
from .dart.api import get_corp_code
from .krx.api import get_krx_all


LOGGER = logging.getLogger(__name__)


class StockCorps:
    """
    mixin KRX market and DART info
    """
    def __init__(self):
        self._corps: Dict[str, KrxCorp] = {}
        self.fetch_info()

    def fetch_info(self) -> None:
        krx_corps = get_krx_all()
        for dart_info in get_corp_code():
            stock_code = dart_info['stock_code']
            if stock_code is not None:
                if stock_code not in krx_corps:
                    # 보통 KRX 에서 상장 폐지 종목은 제외로 아래 warning
                    LOGGER.warning('cannot find dart stock code %s in krx(%s)',
                                   stock_code, dart_info)
                    continue
                krx_corps[stock_code].update_date = dart_info['modify_date']
                krx_corps[stock_code].corp_code = dart_info['corp_code']
                self._corps[stock_code] = krx_corps[stock_code]
    
    def search_by_stock_code(self, stock_code: str):
        return self._corps.get(stock_code)
    
    @property
    def corps(self):
        return self._corps



# -*- coding: utf-8 -*-
import logging


from .krx.api import (
    get_krx_all,
    get_stock_market_list,
    get_trading_halt_list
)

from .dart.api import (
    disclosure_list,
    get_corp_code,
    get_corp_info,
    get_executive_holders,
    get_major_holder_changes,
    fnltt_singl_acnt,
    fnltt_singl_acnt_all
)


from .stock_corps import StockCorps
from .common.types import KrxCorp

__all__ = (
    "StockCorps",
    "KrxCorp",
    "get_krx_all",
    "get_stock_market_list",
    "get_trading_halt_list",
    "disclosure_list",
    "get_corp_code",
    "get_corp_info",
    "get_executive_holders",
    "get_major_holder_changes",
    "fnltt_singl_acnt",
    "fnltt_singl_acnt_all"
)


# Attach a NullHandler to the top level logger by default
# https://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library

logging.getLogger("krxmarket").addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
class APIKeyError(ValueError):
    """
    등록되지 않은 API 키에 의해 발생하는 오류
    """
    def __init__(self, msg='Unregistered API key'):
        super().__init__(msg)


class TemporaryLocked(RuntimeError):
    """
    오픈 API에 등록 되었으나, 일시적으로 사용 중지된 키에 의해 발생하는 오류
    """
    def __init__(self, msg='Temporarily locked'):
        super().__init__(msg)


class NoDataReceived(ValueError):
    """
    조회된 데이터가 없을때 발생하는 오류
    """
    def __init__(self, msg='No data received'):
        super().__init__(msg)


class OverQueryLimit(RuntimeError):
    """
    요청 제한을 초과하였을때 발생하는 오류
    """
    def __init__(self, msg='Over query limit'):
        super().__init__(msg)


class InvalidField(ValueError):
    """
    요청한 필드에 부적절한 값이 있는 경우 발생하는 오류
    """
    def __init__(self, msg='Invalid field'):
        super().__init__(msg)


class ServiceClose(RuntimeError):
    """
    원할한 공시서비스를 위해 오픈 API 서비스가 중지 되었을때 발생하는 오류
    """
    def __init__(self, msg='Open API was closed for web service'):
        super().__init__(msg)


class UnknownError(RuntimeError):
    """
    정의되지 않은 오류
    """
    def __init__(self, msg='Unknown error'):
        super().__init__(msg)


class NotFoundConsolidated(ValueError):
    """
    연결재무제표가 없을때 발생하는 오류
    """
    def __init__(
        self,
        err_msg='Could not find consolidated financial statements'
    ):
        super().__init__(err_msg)



from typing import Literal
import copy
import tempfile

from .status import check_status
from .auth import get_api_key
from .request import api_request, download_xbrl
from ..common.xbrl.xbrl import Xbrl
from ..common.webrequest import request
from ..common.file import read_file_from_zip
from ..common.types import KrxCorp


class Report:
    _DART_URL_ = 'https://dart.fss.or.kr'
    _REPORT_URL_ = _DART_URL_ + '/dsaf001/main.do'
    _DOWNLOAD_URL_ = _DART_URL_ + '/pdf/download/main.do'

    def __init__(self, **kwargs):
        self.rcp_no = (kwargs['rcept_no'] if 'rcept_no' in kwargs
                       else kwargs.get('rcp_no'))
        if self.rcp_no is None:
            raise ValueError('no rcp number in report')
        self.dcm_no = kwargs.get('dcm_no')
        self.info = copy.deepcopy(kwargs)
        """
        info example
        {
            'corp_code': '00126380',
            'corp_name': '삼성전자',
            'stock_code': '005930',
            'corp_cls': 'Y',
            'report_nm': '기타경영사항(자율공시)',
            'rcept_no': '20230214801207',
            'flr_nm': '삼성전자',
            'rcept_dt': '20230214',
            'rm': '유'
        }
        """
        self.xbrl = None

    def load_xbrl(self):
        with tempfile.TemporaryDirectory() as path:
            try:
                file_uri = download_xbrl(path=path, rcept_no=self.rcp_no)
                if len(file_uri) > 0:
                    self.xbrl = Xbrl(file_uri)
            except Exception:
                pass


class SearchResult:
    """ DART 검색결과 정보를 저장하는 클래스"""

    def __init__(self, resp):
        self._page_no = resp['page_no']
        self._page_count = resp['page_count']
        self._total_count = resp['total_count']
        self._total_page = resp['total_page']
        self._report_list = [Report(**x) for x in resp['list']]

    @property
    def page_no(self):
        """ 표시된 페이지 번호 """
        return self._page_no

    @property
    def page_count(self):
        """페이지당 표시할 리포트수"""
        return self._page_count

    @property
    def total_count(self):
        """int: 총 건수"""
        return self._total_count

    @property
    def total_page(self):
        """int: 총 페이지수"""
        return self._total_page

    @property
    def report_list(self):
        """list of Report: 검색된 리포트 리스트"""
        return self._report_list

    def pop(self, index=-1):
        """ 주어진 index 의 리포트를 반환하며, 리스트에서 삭제하는 함수"""
        return self._report_list.pop(index)

    def __len__(self):
        return len(self._report_list)


def get_corp_code() -> list:
    """ DART에 등록되어있는 공시대상회사의 고유번호,회사명,대표자명,종목코드, 최근변경일자 다운로드

    Returns
    -------
    OrderedDict
        고유번호 및 회사 정보
    """
    url = 'https://opendart.fss.or.kr/api/corpCode.xml'
    # Set API KEY
    api_key = get_api_key()
    payload = {'crtfc_key': api_key}

    # Request Download
    file_url = request.download(url=url, payload=payload)
    
    data = read_file_from_zip(file_url, 'CORPCODE.xml')
    if len(data) == 0:
        raise FileNotFoundError('CORPCODE.xml Not Found')
    return data['result']['list']


def get_corp_info(corp_code: str):
    """ 기업 개황 조회

    Parameters
    ----------
    corp_code: str
        공시대상회사의 고유번호(8자리)

    Returns
    -------
    dict
        기업 개황
    """
    path = '/api/company.json'

    return api_request(path=path, corp_code=corp_code)


def get_executive_holders(corp_code: str) -> dict:
    """ 
    임원ㆍ주요주주특정증권등 소유상황보고서 내에 임원ㆍ주요주주 소유보고 
    정보를 제공합니다.

    Parameters
    ----------
    corp_code: str
        공시대상회사의 고유번호(8자리)※ 개발가이드 > 공시정보 > 고유번호 참고
    api_key: str, optional
        DART_API_KEY, 만약 환경설정 DART_API_KEY를 설정한 경우 제공하지 않아도 됨
    Returns
    -------
    dict
        임원ㆍ주요주주 소유보고
    """

    path = '/api/elestock.json'

    return api_request(
        path=path,
        corp_code=corp_code,
    )


def get_major_holder_changes(
    corp_code: str,
) -> dict:
    """ 주식등의 대량보유상황보고서 내에 대량보유 상황보고 정보를 제공합니다.

    Parameters
    ----------
    corp_code: str
        공시대상회사의 고유번호(8자리)※ 개발가이드 > 공시정보 > 고유번호 참고
    api_key: str, optional
        DART_API_KEY, 만약 환경설정 DART_API_KEY를 설정한 경우 제공하지 않아도 됨
    Returns
    -------
    dict
        대량보유 상황보고
    """

    path = '/api/majorstock.json'

    return api_request(
        path=path,
        corp_code=corp_code,
    )


def disclosure_list(
    krx_corp: KrxCorp,
    start_time: str = None,
    end_time: str = None,
    last_reprt_at: Literal['N', 'Y'] = 'N',
    pblntf_ty: str = None,
    pblntf_detail_ty: str = None,
    sort: Literal['date', 'rpt', 'crp'] = 'date',
    sort_mth: Literal['desc', 'asc'] = 'desc',
    page_no: int = 1,
    page_count: int = 10
) -> SearchResult:
    """ 
    개발가이드 -> 공시정보 -> 공시 검색
    Parameters
    ----------
    start_time: str, optional
        검색시작 접수일자(YYYYMMDD), 없으면 종료일(end_de)
    end_time: str, optional
        검색종료 접수일자(YYYYMMDD), 없으면 당일
    last_reprt_at: str, optional
        최종보고서만 검색여부(Y or N), default : N
    pblntf_ty: see report_type.py
    pblntf_detail_ty: see report_type.py
    sort: str, optional
        정렬, {접수일자: date, 회사명: crp, 고서명: rpt}
    sort_mth: str, optional
        오름차순(asc), 내림차순(desc), default : desc
    page_no: int, optional
        페이지 번호(1~n) default : 1
    page_count: int, optional
        페이지당 건수(1~100) 기본값 : 10, default : 100

    Returns
    -------
    dict
        Response data
    """
    url = 'https://opendart.fss.or.kr/api/list.json'

    api_key = get_api_key()

    payload = {
        'crtfc_key': api_key,
        'corp_code': krx_corp.corp_code,
        'bgn_de': start_time,
        'end_de': end_time,
        'last_reprt_at': last_reprt_at,
        'pblntf_ty': pblntf_ty,
        'pblntf_detail_ty': pblntf_detail_ty,
        'corp_cls': krx_corp.short_market_type,
        'sort': sort,
        'sort_mth': sort_mth,
        'page_no': page_no,
        'page_count': page_count
    }

    resp = request.get(url=url, payload=payload)
    dataset = resp.json()
    check_status(**dataset)
    return SearchResult(dataset)


def fnltt_singl_acnt(
    corp_code: str,
    bsns_year: str,
    reprt_code: str
) -> dict:
    """
    상장법인(금융업 제외)이 제출한 정기보고서 내에 XBRL재무제표의 주요계정과목(재무상태표, 손익계산서)을 제공합니다.
    Parameters
    ----------
    corp_code: str
        공시대상회사의 고유번호(8자리)※ 개발가이드 > 공시정보 > 고유번호 참고
    bsns_year: str
        사업연도(4자리) ※ 2015년 이후 부터 정보제공
    reprt_code: str
        1분기보고서:11013, 반기보고서:11012, 3분기보고서:11014, 사업보고서 : 11011
    api_key: str, optional
        DART_API_KEY, 만약 환경설정 DART_API_KEY를 설정한 경우 제공하지 않아도 됨
    Returns
    -------
    dict
        단일회사 주요계정
    """

    return api_request(
        path='/api/fnlttSinglAcnt.json',
        corp_code=corp_code,
        bsns_year=bsns_year,
        reprt_code=reprt_code,
    )


def fnltt_singl_acnt_all(
    corp_code: str,
    bsns_year: str,
    reprt_code: str,
    fs_div: Literal['CFS', 'OFS']
) -> dict:
    """ 
    상장법인(금융업 제외)이 제출한 정기보고서 내에 XBRL재무제표의 모든계정과목을 제공합니다.
    Parameters
    ----------
    corp_code: str
        공시대상회사의 고유번호(8자리)※ 개발가이드 > 공시정보 > 고유번호 참고
    bsns_year: str
        사업연도(4자리) ※ 2015년 이후 부터 정보제공
    reprt_code: str
        1분기보고서:11013, 반기보고서:11012, 3분기보고서:11014, 사업보고서 : 11011
    Returns
    -------
    dict
        단일회사 주요계정
    """
    return api_request(
        path='/api/fnlttSinglAcntAll.json',
        corp_code=corp_code,
        bsns_year=bsns_year,
        reprt_code=reprt_code,
        fs_div=fs_div
    )



from typing import Dict, List
from bs4 import BeautifulSoup

from ..common.webrequest import request
from ..common.types import KrxCorp, MarketType


def get_stock_market_list(market_type: MarketType) -> List[KrxCorp]:
    """
    return corporation list which listed on KRX
    """
    url = 'https://kind.krx.co.kr/corpgeneral/corpList.do'
    referer = 'https://kind.krx.co.kr/corpgeneral/corpList.do' \
              '?method=loadInitPage'
    
    payload = {
        'method': 'download',
        'pageIndex': 1,
        'currentPageSize': 5000,
        'orderMode': 3,
        'orderStat': 'D',
        'searchType': 13,
        'marketType': market_type.value,
        'fiscalYearEnd': 'all',
        'location': 'all',
    }
    
    stock_market_list = []

    resp = request.post(url=url, payload=payload, referer=referer)
    html = BeautifulSoup(resp.text, 'html.parser')
    rows = html.find_all('tr')

    for row in rows:
        cols = row.find_all('td')
        if len(cols) > 0:
            stock_market_list.append(
                KrxCorp(
                    sectors=cols[2].text.strip(),
                    products=cols[3].text.strip(),
                    market_type=market_type,
                    listed_date=cols[4].text.strip(),
                    name=cols[0].text.strip(),
                    stock_code=cols[1].text.strip()
                )
            )

    return stock_market_list


def _market_type_to_int(market_type: MarketType) -> int:
    if market_type == MarketType.KOSPI:
        return 1
    elif market_type == MarketType.KOSDAQ:
        return 2
    elif market_type == MarketType.KONEX:
        return 6
    raise TypeError('cannot find market type')


def get_trading_halt_list(market_type: MarketType) -> dict:
    """ 
    return dictionary of halted corps, key: stock code, value: issue
    """
    url = 'https://kind.krx.co.kr/investwarn/tradinghaltissue.do'
    referer = 'https://kind.krx.co.kr/investwarn/' \
              'tradinghaltissue.do?method=searchTradingHaltIssueMain'

    payload = {
        'method': 'searchTradingHaltIssueSub',
        'currentPageSize': 3000,
        'pageIndex': 1,
        'searchMode': '',
        'searchCodeType': '',
        'searchCorpName': '',
        'forward': 'tradinghaltissue_down',
        'paxreq': '',
        'outsvcno': '',
        'marketType': _market_type_to_int(market_type),
        'repIsuSrtCd': '',
    }

    trading_halt_dict = {}

    resp = request.post(url=url, payload=payload, referer=referer)
    html = BeautifulSoup(resp.text, 'html.parser')
    rows = html.find_all('tr')

    for row in rows:
        cols = row.find_all('td')
        if len(cols) > 0:
            trading_halt_dict[cols[2].text.strip()] = cols[3].text.strip()

    return trading_halt_dict


def get_krx_all(exclude_halt=True) -> Dict[str, KrxCorp]:
    """
    return dictionary(key: stock_code, value: KrxInfo)
    """
    market_types = [MarketType.KOSPI, MarketType.KOSDAQ, MarketType.KONEX]
    halt_corps = {}
    corps: Dict[str, KrxCorp] = {}
    for market_type in market_types:
        halt_corps = {**get_trading_halt_list(market_type)}

    for market_type in market_types:
        market_list = get_stock_market_list(market_type)
        for corp in market_list:
            if corp.stock_code in halt_corps.keys():
                corp.is_halt = True
                corp.halt_reason = halt_corps[corp.stock_code]

            if not corp.is_halt or not exclude_halt:
                corps[corp.stock_code] = corp

    return corps


from krxmarket.stock_corps import StockCorps
from krxmarket.performance import get_performance


def test_performance_all():
    corps = StockCorps().corps

    for stock_code, corp in corps.items():
        get_performance(corp.corp_code, 2016)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from re import search

from setuptools import setup


def get_version():
    with open('krxmarket/version.py') as version_file:
        return search(r"""__version__\s+=\s+(['"])(?P<version>.+?)\1""",
                      version_file.read()).group('version')


setup(
    name="krxmarket",
    version=get_version()
)



import tempfile
import zipfile
import re
import os
import xmltodict
import pathlib


def read_file_from_zip(download_uri: str, filename: str):
    with tempfile.TemporaryDirectory() as path:
        with zipfile.ZipFile(download_uri, 'r') as zip_ref:
            zip_ref.extractall(path=path)

        for root, _, files in os.walk(path):
            for file in files:
                if re.search(filename, file):
                    with open(os.path.join(root, file), encoding='utf8') as f:
                        content = xmltodict.parse(f.read())
                        return content
    return {}


def unzip_xbrl(dest_directory: str, download_uri: str):
    with zipfile.ZipFile(download_uri, 'r') as zip_ref:
        zip_ref.extractall(path=dest_directory)
    for root, _, files in os.walk(dest_directory):
        for file in files:
            if pathlib.Path(os.path.join(root, file)).suffix == '.xbrl':
                return os.path.join(root, file)
    return ''




pblntf_ty = {
    "A": "정기공시",
    "B": "주요사항보고",
    "C": "발행공시",
    "D": "지분공시",
    "E": "기타공시",
    "F": "외부감사관련",
    "G": "펀드공시",
    "H": "자산유동화",
    "I": "거래소공시",
    "J": "공정위공시"
}

pblntf_detail_ty = {
    "A001": "사업보고서",
    "A002": "반기보고서",
    "A003": "분기보고서",
    "A004": "등록법인결산서류(자본시장법이전)",
    "A005": "소액공모법인결산서류",
    "B001": "주요사항보고서",
    "B002": "주요경영사항신고(자본시장법 이전)",
    "B003": "최대주주등과의거래신고(자본시장법 이전)",
    "C001": "증권신고(지분증권)",
    "C002": "증권신고(채무증권)",
    "C003": "증권신고(파생결합증권)",
    "C004": "증권신고(합병등)",
    "C005": "증권신고(기타)",
    "C006": "소액공모(지분증권)",
    "C007": "소액공모(채무증권)",
    "C008": "소액공모(파생결합증권)",
    "C009": "소액공모(합병등)",
    "C010": "소액공모(기타)",
    "C011": "호가중개시스템을통한소액매출",
    "D001": "주식등의대량보유상황보고서",
    "D002": "임원ㆍ주요주주특정증권등소유상황보고서",
    "D003": "의결권대리행사권유",
    "D004": "공개매수",
    "E001": "자기주식취득/처분",
    "E002": "신탁계약체결/해지",
    "E003": "합병등종료보고서",
    "E004": "주식매수선택권부여에관한신고",
    "E005": "사외이사에관한신고",
    "E006": "주주총회소집공고",
    "E007": "시장조성/안정조작",
    "E008": "합병등신고서(자본시장법 이전)",
    "E009": "금융위등록/취소(자본시장법 이전)",
    "F001": "감사보고서",
    "F002": "연결감사보고서",
    "F003": "결합감사보고서",
    "F004": "회계법인사업보고서",
    "F005": "감사전재무제표미제출신고서",
    "G001": "증권신고(집합투자증권-신탁형)",
    "G002": "증권신고(집합투자증권-회사형)",
    "G003": "증권신고(집합투자증권-합병)",
    "H001": "자산유동화계획/양도등록",
    "H002": "사업/반기/분기보고서",
    "H003": "증권신고(유동화증권등)",
    "H004": "채권유동화계획/양도등록",
    "H005": "수시보고",
    "H006": "주요사항보고서",
    "I001": "수시공시",
    "I002": "공정공시",
    "I003": "시장조치/안내",
    "I004": "지분공시",
    "I005": "증권투자회사",
    "I006": "채권공시",
    "J001": "대규모내부거래관련",
    "J002": "대규모내부거래관련(구)",
    "J004": "기업집단현황공시",
    "J005": "비상장회사중요사항공시",
    "J006": "기타공정위공시"
}
